# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from coveriteam.language import CoVeriLangException
import os
from zipfile import ZipFile, ZIP_DEFLATED, ZipInfo
import re
import urllib.request
import requests
from pathlib import Path
import shutil
import sys

LOG_DIR = Path.cwd() / "cvt-output"
TOOL_OUTPUT_FILE = "output.txt"
INPUT_FILE_DIR = f'{Path(__file__).parent.resolve() / "artifactlibrary/"}/'
ACTOR_CONFIG_PATH = ""

CVT_DEBUG_LEVEL = 15


def set_cache_directories(d=None):
    global INSTALL_DIR, ARCHIVE_DOWNLOAD_PATH, TOOL_INFO_DOWNLOAD_PATH
    if d:
        cache_dir = d
    elif os.getenv("XDG_CACHE_HOME"):
        cache_dir = Path(os.getenv("XDG_CACHE_HOME")) / "coveriteam"
    else:
        cache_dir = Path.home() / ".cache" / "coveriteam"
    INSTALL_DIR = cache_dir / "tools"
    ARCHIVE_DOWNLOAD_PATH = cache_dir / "archives"
    TOOL_INFO_DOWNLOAD_PATH = cache_dir / "toolinfocache"
    sys.path.append(str(TOOL_INFO_DOWNLOAD_PATH))
    create_cache_directories()


def set_cache_update(flag):
    global UPDATE_CACHE
    UPDATE_CACHE = flag


def create_cache_directories():
    # Create directories and set path.
    if not ARCHIVE_DOWNLOAD_PATH.is_dir():
        ARCHIVE_DOWNLOAD_PATH.mkdir(parents=True)
    if not TOOL_INFO_DOWNLOAD_PATH.is_dir():
        TOOL_INFO_DOWNLOAD_PATH.mkdir(parents=True)

    if not INSTALL_DIR.is_dir():
        INSTALL_DIR.mkdir(parents=True)


def get_INSTALL_DIR():
    return INSTALL_DIR


def get_ARCHIVE_DOWNLOAD_PATH():
    return ARCHIVE_DOWNLOAD_PATH


def get_TOOL_INFO_DOWNLOAD_PATH():
    return TOOL_INFO_DOWNLOAD_PATH


def is_url(path_or_url):
    return "://" in path_or_url or path_or_url.startswith("file:")


def make_url(path_or_url):
    """Make a URL from a string which is either a URL or a local path,
    by adding "file:" if necessary.
    """
    if not is_url(path_or_url):
        return "file:" + urllib.request.pathname2url(path_or_url)
    return path_or_url


def download_if_needed(location, target):
    url = make_url(location)
    headers = {"User-Agent": "Mozilla"}
    etag_path = Path(str(target) + ".etag")
    if etag_path.is_file() and target.is_file():
        etag = etag_path.open("r").read()
        headers["If-None-Match"] = etag

    try:
        response = requests.get(url, allow_redirects=True, headers=headers)
        if response.status_code == 304:
            return False
        if response.status_code != 200:
            msg = (
                "Couldn't download tool archive from: %s. Server returned the code: %s"
                % (str(url), response.status_code)
            )
            raise CoVeriLangException(msg)
        with target.open("wb") as out_file:
            out_file.write(response.content)
        # write the eta tag
        with Path(str(target) + ".etag").open("w") as f:
            f.write(response.headers["etag"])
        return True
    except requests.ConnectionError:
        print("No network access. Running in offline mode.")
        return False


def create_archive(dirname, archive_path):
    with ZipFile(archive_path, "w", ZIP_DEFLATED) as zipf:
        for root, _dirs, files in os.walk(dirname):
            for f in files:
                filepath = os.path.join(root, f)
                zipf.write(filepath, os.path.relpath(filepath, dirname))


def unzip(archive, target_dir):
    if target_dir.is_dir():
        shutil.rmtree(target_dir)

    with ZipFile(archive, "r") as zipfile:
        top_folder = INSTALL_DIR / zipfile.filelist[0].filename.split("/")[0]
        # Not to use extract all as it does not preserves the permission for executable files.
        # See: https://bugs.python.org/issue15795
        # See https://stackoverflow.com/questions/39296101/python-zipfile-removes-execute-permissions-from-binaries
        for member in zipfile.namelist():
            if not isinstance(member, ZipInfo):
                member = zipfile.getinfo(member)
            extracted_file = zipfile.extract(member, INSTALL_DIR)
            # This takes first two bytes from four bytes.
            attr = member.external_attr >> 16
            if attr != 0:
                os.chmod(extracted_file, attr)
        top_folder.rename(target_dir)


def filter_dict(d, d1):
    return {k: d[k] for k in d1.keys()}


def str_dict(d):
    return {k: str(d[k]) for k in d.keys()}


def dict_clash(d1, d2):
    """
    This function checks if there is a key present in both dictionaries
    whose values are different.
    """
    for k in d1.keys():
        if k in d2.keys() and d1[k] != d2[k]:
            return True

    return False


def rename_dict(d, renaming_map):
    return {(renaming_map.get(k, None) or k): d[k] for k in d.keys()}


def collect_variables(exp):
    regex_isinstance = r"(?<=isinstance\()\S+(?=,)"
    regex_in = r"\w+(?= in \[)"
    regex = regex_isinstance + "|" + regex_in
    names = re.findall(regex, exp)

    return names


def get_additional_paths_for_container_config():
    base_dir = Path(__file__).parent.parent.resolve()
    paths = [str(base_dir / "lib"), str(base_dir / "coveriteam" / "toolconfigs")]
    return paths



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from benchexec import tooladapter
from benchexec.model import cmdline_for_run, load_tool_info
from benchexec.runexecutor import RunExecutor
import benchexec.containerexecutor as containerexecutor
from benchexec.test_benchmark_definition import DummyConfig
from coveriteam.language.actor import Actor
from coveriteam.language.actorconfig import ActorConfig
from coveriteam.util import (
    TOOL_OUTPUT_FILE,
    str_dict,
    get_additional_paths_for_container_config,
    get_TOOL_INFO_DOWNLOAD_PATH,
)
import argparse
from pathlib import Path
import os
import uuid
from xml.etree import ElementTree
from coveriteam.language import CoVeriLangException
from benchexec.container import DIR_HIDDEN, DIR_OVERLAY, DIR_READ_ONLY, DIR_FULL_ACCESS
from string import Template
from coveriteam.language.artifact import AtomicActorDefinition
import logging


def get_task_metadata(program_path):
    """
    This function returns the task options to be passed to the tool info module
    to create a command.
    At the moment we only have two kinds of programs: C and Java.
    """
    d = {}
    if isinstance(program_path, str) and Path(program_path).suffix in [".i", ".c"]:
        if not Actor.data_model:
            raise CoVeriLangException(
                "Data model missing. Please provide a data model."
            )
        # TODO Actor is not a nice place to store data model. Think about it.
        d = {"data_model": Actor.data_model, "language": "C"}
    else:
        if Actor.data_model:
            raise CoVeriLangException("Data model is not allowed in Java verification.")
        d = {"language": "Java"}
    return d


class AtomicActor(Actor):
    def __init__(self, path, version=None):
        if isinstance(path, AtomicActorDefinition):
            path = path.path
        self.__config = ActorConfig(path, version)

    def name(self):
        return self.__config.actor_name

    def log_dir(self):
        # actor execution id is for the complete execution of an actor -- atomic or composite
        # atomic execution id is for this specific atomic actor.
        return (
            Actor.get_top_actor_execution_dir()
            / self.name()
            / self._atomic_execution_id
        )

    def log_file(self):
        return self.log_dir() / TOOL_OUTPUT_FILE

    def _get_relative_path_to_tool(self, path):
        return os.path.relpath(path, self.__config.tool_dir) if path else ""

    def print_version(self):
        cwd = os.getcwd()
        os.chdir(self.__config.tool_dir)
        self.__set_directory_modes({})

        tool_name = self.__config.tool_name or self.name()
        tool_info, self._tool = load_tool_info(
            tool_name, self.__create_config_for_container_execution()
        )
        tool_locator = tooladapter.CURRENT_BASETOOL.ToolLocator(
            use_path=True, use_current=True
        )
        version = self._tool.version(self._tool.executable(tool_locator))
        print(self._tool.name() + " " + version)
        os.chdir(cwd)

    def act(self, **kwargs):
        # Generate atomic execution id and then call the act method of the super class.
        self._atomic_execution_id = str(uuid.uuid4())
        self.__set_directory_modes(kwargs)
        res = super().act(**kwargs)
        self.gen_xml_elem(kwargs, res)

        return res

    def _act(self, **kwargs):
        args = self._prepare_args(**kwargs)
        d = self._get_arg_substitutions(**kwargs)
        options = [Template(o).safe_substitute(**d) for o in self.__config.options]
        self._run_tool(*args, options)
        try:
            res = self._extract_result()
            self._tool.close()
            return res
        except UnboundLocalError:
            msg = "The execution of the actor {} did not produce the expected result".format(
                self.name()
            )
            msg += "More information can be found in the logfile produced by the tool: {}".format(
                self.log_file()
            )
            raise CoVeriLangException(msg)

    def _run_tool(self, program_path, property_path, additional_options=[], options=[]):
        # Change directory to tool's directory
        cwd = os.getcwd()
        os.chdir(self.__config.tool_dir)

        task_options = get_task_metadata(program_path)
        if isinstance(program_path, str):
            program_path = [self._get_relative_path_to_tool(program_path)]
        elif isinstance(program_path, list):
            program_path = [self._get_relative_path_to_tool(p) for p in program_path]

        property_path = self._get_relative_path_to_tool(property_path)

        tool_name = self.__config.tool_name or self.name()

        tool_info, self._tool = load_tool_info(
            tool_name, self.__create_config_for_container_execution()
        )
        lims_for_exec = {
            "softtimelimit": self.__config.reslim.get("timelimit"),
            "memlimit": self.__config.reslim.get("memlimit"),
        }
        resource_limits = tooladapter.CURRENT_BASETOOL.ResourceLimits(
            self.__config.reslim.get("timelimit"),
            self.__config.reslim.get("timelimit"),
            None,
            self.__config.reslim.get("memlimit"),
            self.__config.reslim.get("cpuCores"),
        )
        tool_locator = tooladapter.CURRENT_BASETOOL.ToolLocator(
            use_path=True, use_current=True
        )
        tool_executable = self._tool.executable(tool_locator)
        # TODO This is bad. It has to change. cmd should not be a part of the actor.
        # But we need it to extract result since PR 592 in benchexec.
        self._cmd = cmdline_for_run(
            self._tool,
            tool_executable,
            options + additional_options,
            program_path,
            None,
            property_path,
            task_options,
            resource_limits,
        )

        self.measurements = RunExecutor(dir_modes=self._dir_modes).execute_run(
            self._cmd,
            str(self.log_file().resolve()),
            output_dir=str(self.log_dir().resolve()),
            result_files_patterns=self._result_files_patterns,
            workingDir=self._tool.working_directory(tool_executable),
            environments=self._tool.environment(tool_executable),
            **lims_for_exec
        )
        terminationreason = self.measurements.get("terminationreason")
        if terminationreason:
            logging.warning(
                "The actor %s was terminated by BenchExec. Termination reason: %s."
                " Possibly it did not produce the expected result."
                % (self.name(), terminationreason)
            )

        # Change back to the original directory
        os.chdir(cwd)

    def gen_xml_elem(self, inputs, outputs):
        super().gen_xml_elem(inputs, outputs)
        data = self.get_measurements_data_for_xml()
        self.xml_elem.append(ElementTree.Element("measurements", str_dict(data)))
        tool_output_elem = ElementTree.Element("tool_output")
        tool_output_elem.text = str(Actor._get_relative_path_to_actor(self.log_file()))
        self.xml_elem.append(tool_output_elem)

    def get_measurements_data_for_xml(self):
        data_filter = ["cputime", "walltime", "memory"]
        data = {k: self.measurements[k] for k in data_filter}
        return str_dict(data)

    def __set_directory_modes(self, inputs):
        # The default directory modes taken from container executor.
        self._dir_modes = {
            "/": DIR_READ_ONLY,
            "/run": DIR_HIDDEN,
            "/tmp": DIR_HIDDEN,  # noqa S108
        }
        # Update the default with the /sys and /home as hidden.
        self._dir_modes["/sys"] = DIR_HIDDEN
        self._dir_modes["/home"] = DIR_HIDDEN
        self._dir_modes[self.__config.tool_dir] = DIR_OVERLAY
        self._dir_modes[str(get_TOOL_INFO_DOWNLOAD_PATH())] = DIR_OVERLAY

        for v in inputs.values():
            if isinstance(v.path, str):
                p = str(Path(v.path).parent.resolve())
                self._dir_modes[p] = DIR_OVERLAY
            elif isinstance(v.path, list):
                for path in v.path:
                    p = str(Path(path).parent.resolve())
                    self._dir_modes[p] = DIR_OVERLAY

    def __create_config_for_container_execution(self):
        try:
            if Actor.trust_tool_info:
                return DummyConfig
        except AttributeError:
            # We use the standard container config if the field doesn't exist.
            pass

        parser = argparse.ArgumentParser()
        containerexecutor.add_basic_container_args(parser)
        containerexecutor.add_container_output_args(parser)
        mp = {
            DIR_HIDDEN: "--hidden-dir",
            DIR_OVERLAY: "--overlay-dir",
            DIR_READ_ONLY: "--read-only-dir",
            DIR_FULL_ACCESS: "--full-access-dir",
        }
        args = []
        for p in get_additional_paths_for_container_config():
            args += ["--full-access-dir", p]
        for k, v in self._dir_modes.items():
            args += [mp[v], k]
        config = parser.parse_args(args)
        config.container = True
        return config

    def _get_arg_substitutions(self, **kwargs):
        return {}



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

# Generated from CoVeriLang.g4 by ANTLR 4.7.2
from pathlib import Path
from coveriteam.interpreter import get_parsed_tree
from coveriteam.parser.CoVeriLangParser import CoVeriLangParser
from coveriteam.parser.CoVeriLangVisitor import CoVeriLangVisitor


def remove_quotes(s):
    return s[1:-1]


# This class defines a complete listener for a parse tree produced by CoVeriLangParser.
class FileCollector(CoVeriLangVisitor):

    files = []

    def __init__(self, path):
        # TODO don't know a better way to do it.
        # I need the cvt path to resolve the file paths.
        self.__cvt_path = Path(path).parent.resolve()
        super().__init__()

    # Visit a parse tree produced by CoVeriLangParser#spec_stmt.
    def visitSpec_stmt(self, ctx: CoVeriLangParser.Spec_stmtContext):
        a = ctx.assignable()
        if a.STRING():
            self.files += [remove_quotes(a.STRING().getText())]
        else:
            self.visitAssignable(a)

    # Visit a parse tree produced by CoVeriLangParser#Atomic.
    def visitAtomic(self, ctx: CoVeriLangParser.AtomicContext):
        if ctx.STRING():
            actordef = remove_quotes(ctx.STRING().getText())
            self.files += [actordef]

    # Visit a parse tree produced by CoVeriLangParser#artifact.
    def visitArtifact(self, ctx: CoVeriLangParser.ArtifactContext):
        if ctx.STRING():
            path = remove_quotes(ctx.STRING().getText())
            self.files += [path]


def collect_files(path_str):
    tree = get_parsed_tree(path_str)

    visitor = FileCollector(path_str)
    visitor.visit(tree)
    return visitor.files



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from coveriteam.language.artifact import (
    CProgram,
    Condition,
    Witness,
    Specification,
    TestGoal,
    TestSpecification,
    TestSuite,
    AtomicActorDefinition,
)
from coveriteam.language.actor import Actor, Instrumentor, Reducer, Transformer
from coveriteam.language.atomicactor import AtomicActor
from coveriteam.util import create_archive
from benchexec import tooladapter
import os
import logging


class WitnessInstrumentor(Instrumentor, AtomicActor):
    _input_artifacts = {"program": CProgram, "witness": Witness}
    _output_artifacts = {"program": CProgram}
    _result_files_patterns = ["**/*.c"]

    def _prepare_args(self, program, witness):
        return [program.path, witness.path]

    def _extract_result(self):
        # extract result
        for file in self.log_dir().glob("**/*.c"):
            instrumentedProgram = CProgram(file)

        return {"program": instrumentedProgram}


class WitnessToTest(AtomicActor):
    _input_artifacts = {"program": CProgram, "spec": Specification, "witness": Witness}
    _output_artifacts = {"test_suite": TestSuite}
    _result_files_patterns = ["**/*.xml"]

    def _prepare_args(self, program, spec, witness):
        options = ["-witness", self._get_relative_path_to_tool(witness.path)]
        return [program.path, spec.path, options]

    def _extract_result(self):
        # We assume that the test generator will succeed and create a directory
        # containing metadata.xml. This directory is the test suite.
        for file in self.log_dir().glob("**/metadata.xml"):
            testSuite = TestSuite(os.path.dirname(file))

        return {"test_suite": testSuite}


class TestCriterionInstrumentor(Instrumentor, AtomicActor):
    _input_artifacts = {"program": CProgram, "test_spec": TestSpecification}
    _output_artifacts = {"program": CProgram}
    _result_files_patterns = ["**/*.c"]

    def _prepare_args(self, program, test_spec):
        return [program.path, test_spec.path]

    def _extract_result(self):
        # extract result
        for file in self.log_dir().glob("**/*.c"):
            instrumentedProgram = CProgram(file)

        return {"program": instrumentedProgram}


class TestGoalPruner(Reducer, AtomicActor):
    _input_artifacts = {
        "program": CProgram,
        "test_spec": TestSpecification,
        "covered_goals": TestGoal,
    }
    _output_artifacts = {"program": CProgram}
    _result_files_patterns = ["**/*.c"]

    def _prepare_args(self, program, test_spec, covered_goals):
        options = []
        if covered_goals.path:
            options += [
                "--covered-labels",
                self._get_relative_path_to_tool(covered_goals.path),
            ]

        return [program.path, test_spec.path, options]

    def _extract_result(self):
        # extract result
        for file in self.log_dir().glob("**/reduced.c"):
            prunedProgram = CProgram(file)

        return {"program": prunedProgram}


class TestGoalAnnotator(Reducer, AtomicActor):
    _input_artifacts = {
        "program": CProgram,
        "test_spec": TestSpecification,
        "covered_goals": TestGoal,
    }
    _output_artifacts = {"program": CProgram}
    _result_files_patterns = ["**/*.c"]

    def _prepare_args(self, program, test_spec, covered_goals):
        options = []
        if covered_goals.path:
            options += [
                "--covered-labels",
                self._get_relative_path_to_tool(covered_goals.path),
            ]
        return [program.path, test_spec.path, options]

    def _extract_result(self):
        # extract result
        for file in self.log_dir().glob("**/reduced.c"):
            annotatedProgram = CProgram(file)

        return {"program": annotatedProgram}


class TestGoalExtractor(Transformer, AtomicActor):
    _input_artifacts = {
        "program": CProgram,
        "test_spec": TestSpecification,
        "test_suite": TestSuite,
    }
    _output_artifacts = {"extracted_goals": TestGoal}
    _result_files_patterns = ["**/*.txt"]

    def _prepare_args(self, program, test_spec, test_suite):
        testzip = os.path.join(os.path.dirname(test_suite.path), "test_suite.zip")
        create_archive(test_suite.path, testzip)
        testzip = self._get_relative_path_to_tool(testzip)
        options = ["--test-suite", testzip]
        return [program.path, test_spec.path, options]

    def _extract_result(self):
        # extract result
        for file in self.log_dir().glob("**/covered_goals.txt"):
            extracted_goals = TestGoal(file)

        return {"extracted_goals": extracted_goals}


class CMCReducer(Transformer, AtomicActor):
    _input_artifacts = {"program": CProgram, "condition": Condition}
    _output_artifacts = {"program": CProgram}
    _result_files_patterns = ["**/*.c"]

    def _prepare_args(self, program, condition):
        options_assm_file = [
            "-setprop",
            "residualprogram.assumptionFile="
            + self._get_relative_path_to_tool(condition.path),
        ]
        options_assm_automaton = [
            "-setprop",
            "AssumptionAutomaton.cpa.automaton.inputFile="
            + self._get_relative_path_to_tool(condition.path),
        ]
        options = options_assm_automaton + options_assm_file
        return [program.path, "", options]

    def _extract_result(self):
        # extract result
        for file in self.log_dir().glob("**/*.c"):
            reducedProgram = CProgram(file)

        return {"program": reducedProgram}


class AlgorithmSelector(AtomicActor):
    _input_artifacts = {"program": CProgram, "spec": Specification}
    _output_artifacts = {"actordef": AtomicActorDefinition}
    _result_files_patterns = []

    def _prepare_args(self, program, spec):
        return [program.path, spec.path]

    def _extract_result(self):
        # TODO this could be put in a pattern
        try:
            with open(self.log_file(), "rt", errors="ignore") as outputFile:
                output = outputFile.readlines()
                # first 6 lines are for logging, rest is output of subprocess, see runexecutor.py for details
                output = output[6:]
        except IOError as e:
            logging.warning("Cannot read log file: %s", e.strerror)
            output = []

        exit_code = self.measurements.get("exitcode")
        output = tooladapter.CURRENT_BASETOOL.RunOutput(output)
        run = tooladapter.CURRENT_BASETOOL.Run(self._cmd, exit_code, output, None)

        actordef = AtomicActorDefinition(self._tool.determine_result(run))
        return {"actordef": actordef}


class DynamicActor(Actor):
    # TODO find a place to put this actor.
    # It is neither an atomic not a composite actor.

    def __init__(self, actorkind):
        self.actor_to_execute_class = actorkind
        self._input_artifacts = self.actor_to_execute_class._input_artifacts.copy()
        # Additionally add the actor definition.
        self._input_artifacts.update(actordef=AtomicActorDefinition)
        self._output_artifacts = self.actor_to_execute_class._output_artifacts.copy()

    def name(self):
        return self.actor_to_execute_class.get_actor_kind()

    def _act(self, actordef, **kwargs):
        """
        This should first create an actor based on the actor definition, and then call its _act.
        This should suffice.
        This makes me think it is a composite actor.
        """
        actor_to_execute = self.actor_to_execute_class(actordef.path)
        return actor_to_execute.act(**kwargs)


class ConditionExtractor(Transformer, AtomicActor):
    """
    An actor that takes a C program and a list of coveread goals and generates a condition
    automaton from it. A condition can be used in combination with CPAchecker's reducer.
    """

    _input_artifacts = {
        "program": CProgram,
        "covered_goals": TestGoal,
    }
    _output_artifacts = {"condition": Condition}
    _result_files_patterns = ["**/AssumptionAutomaton.txt"]

    def _prepare_args(self, program, covered_goals):
        options_inputfile = [
            "-setprop",
            "conditional_testing.inputfile="
            + self._get_relative_path_to_tool(covered_goals.path),
        ]

        return [program.path, "", options_inputfile]

    def _extract_result(self):
        for result_file in self.log_dir().glob("**/AssumptionAutomaton.txt"):
            condition = Condition(result_file)

            return {"condition": condition}
        return {"condition": None}



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from coveriteam.language.actor import Actor

# TODO find a better way to do this
from benchexec.result import RESULT_CLASS_FALSE  # noqa: F401,F403,E261
from benchexec.result import RESULT_CLASS_TRUE  # noqa: F401,F403,E261
from coveriteam.language.artifact import *  # noqa: F401,F403,E261
from coveriteam.language import CoVeriLangException
from coveriteam.util import filter_dict, dict_clash, collect_variables
from xml.etree import ElementTree
from textwrap import indent


def infer_types(exp):
    d = {}
    # TODO At the moment it is rudimentory. Putting Artifact for everything
    for name in collect_variables(exp):
        if not name.startswith("RESULT_CLASS"):
            d[name] = Artifact

    return d


class CompositeActor(Actor):
    def name(self):
        try:
            return self._name
        except AttributeError:
            return type(self).__name__

    def set_name(self, name):
        # This is only in the composite actor.
        # The names of the atomic actors come from the actor definition.
        self._name = name

    def forward_artifacts(self, available, renaming, required):
        """
        This function does two things:
        1) Selects the artifacts to be passed to a subcomponent from the available
           artifacts.
        2) Applies renaming if neccessary.
        """
        # TODO at the moment renaming is not considered
        return filter_dict(available, required)


def set_composite_actor_name(actor, name):
    if isinstance(actor, CompositeActor):
        actor.set_name(name)
    else:
        raise CoVeriLangException(
            "Setting name is only supported for composite actors."
        )


class Sequence(CompositeActor):
    def __init__(self, a1: Actor, a2: Actor):
        self.first = a1
        self.second = a2
        self._input_artifacts = self._infer_input_type(a1, a2)
        self._output_artifacts = a2.get_output_artifacts().copy()

    def _act(self, **kwargs):
        kwargs1 = self.forward_artifacts(kwargs, {}, self.first.get_input_artifacts())
        res1 = self.first.act(**kwargs1)
        # This is ugly, but it works
        new_args = filter_dict({**kwargs, **res1}, self.second.get_input_artifacts())
        ret_val = self.second.act(**new_args)
        self.gen_xml_elem(kwargs, ret_val)
        return ret_val

    def __type_check(self):
        a1 = self.a1
        a2 = self.a2
        # At the moment the type checking for sequence is suspended, as the rules have changed.
        # TODO need to check covariance here.
        # TODO This will not check the item but only the names.
        # This should be changed if we do proper type inference.
        """if not a2.get_input_artifacts().items() <= a1.get_output_artifacts().items():"""
        if not a2.get_input_artifacts().keys() <= a1.get_output_artifacts().keys():
            raise CoVeriLangException(
                "Components {} \n and \n {} cannot be composed in sequence."
                "Type check failed.".format(a1, a2)
            )

    def __str__(self):
        return (
            "\nSEQUENCE"
            + self.__get_actor_type_str__()
            + "\n"
            + indent(str(self.first), "\t")
            + "\n"
            + indent(str(self.second), "\t")
        )

    def gen_xml_elem(self, inputs, outputs):
        super().gen_xml_elem(inputs, outputs)
        first = ElementTree.Element("first")
        first.append(self.first.xml_elem)
        second = ElementTree.Element("second")
        second.append(self.second.xml_elem)

        self.xml_elem.append(first)
        self.xml_elem.append(second)

    def _infer_input_type(self, a1, a2):
        ip1 = a1.get_input_artifacts().copy()
        ip2 = a2.get_input_artifacts().copy()
        op1 = a1.get_output_artifacts().copy()

        # TODO check the name clash with two different types

        # First do ip2 -op1, i.e, get the artifacts which cannot be provided
        # by the first actor.
        for k in op1.keys():
            ip2.pop(k, None)
        for k in ip1.keys():
            ip2.pop(k, None)

        return {**ip1, **ip2}


class ITE(CompositeActor):
    """Concerns faced:
    1) the condition might need a variable which is not required by the composing actors.
    -- To avoid this at the moment it is allowd that the inputs could be more than
    required by the composing actors."""

    def __init__(self, cond, a1, a2=None):
        self.__type_check(a1, a2)
        self.cond = cond
        self.first = a1
        self.second = a2
        self._input_artifacts = a1.get_input_artifacts().copy()
        self._output_artifacts = a1.get_output_artifacts().copy()
        self._add_types_from_cond(cond)

    def _act(self, **kwargs):
        # Tried using ast.literal_eval but couldn't use it. Still check it once more
        # noqa because can't pass kwargs to ast.literal_eval
        cond_val = eval(self.cond, globals(), {**kwargs})  # noqa: S307
        if cond_val:
            args_to_pass = self.forward_artifacts(
                kwargs, {}, self.first.get_input_artifacts()
            )
            ret_val = self.first.act(**args_to_pass)
        elif self.second:
            # TODO check if this and the one in if condition are same
            args_to_pass = self.forward_artifacts(
                kwargs, {}, self.second.get_input_artifacts()
            )
            ret_val = self.second.act(**args_to_pass)
        else:
            # In this case the values from the inputs are forwarded.
            ret_val = filter_dict(kwargs, self._output_artifacts)

        self.gen_xml_elem(kwargs, ret_val, cond_val)
        return ret_val

    def __type_check(self, a1, a2):
        # TODO need to check covariance here.
        if not a2:
            if not (
                a1.get_output_artifacts().items() <= a1.get_input_artifacts().items()
            ):
                # In this case we still would like to check if the inputs could be forwarded.
                raise CoVeriLangException(
                    "Type check failed for ITE composition."
                    "Second component is needed. All the outputs cannot be assigned from input."
                )
        elif (
            a1.get_input_artifacts() != a2.get_input_artifacts()
            or a1.get_output_artifacts() != a2.get_output_artifacts()
        ):
            raise CoVeriLangException(
                "Components \n {} \n and \n {} cannot be composed in choice."
                "Type check failed.".format(a1, a2)
            )

    def __str__(self):
        return (
            "\nITE"
            + self.__get_actor_type_str__()
            + "\n"
            + indent(str(self.first), "\t")
            + "\n"
            + indent(str(self.second), "\t")
        )

    def gen_xml_elem(self, inputs, outputs, cond_val):
        super().gen_xml_elem(inputs, outputs)
        cond_elem = ElementTree.Element("condition")
        cond_elem.text = str(cond_val)
        self.xml_elem.append(cond_elem)
        if cond_val:
            self.xml_elem.append(self.first.xml_elem)
        elif self.second:
            self.xml_elem.append(self.second.xml_elem)

    def _add_types_from_cond(self, cond):
        t = infer_types(cond)
        self._input_artifacts = {**t, **self._input_artifacts}


class Iterative(CompositeActor):

    # I think every can be run iteratively. Two cases arise which need to be handled:
    # 1) What if all the outputs are disregarded - It is still a valid composition
    # 2) if outputs are to be fed to inputs but have different names
    def __init__(self, terminationCondition, a: Actor):
        self.__type_check(a)
        self.actor = a
        self.terminationCondition = terminationCondition
        self._input_artifacts = a.get_input_artifacts()
        self._output_artifacts = a.get_output_artifacts()
        # TODO This is making it stateful, might create a problem later.
        self._iteration_count = 0
        self.__child_xml_elems = []
        # TODO Put it in type check too
        self.__artifacts_to_accumulate = (
            a.get_output_artifacts().keys() - a.get_input_artifacts().keys()
        )
        self.__accumulated_artifacts = {}

    def _act(self, **kwargs):
        # TODO Maybe it is a better idea to keep all the objects as they go along.
        # This might solve the xml generation issues.
        res = self.actor.act(**kwargs)
        self.__do_step_iteration_post_process(kwargs, res)
        # BEWARE!! to think if it is OK. This is out of syn with the paper.
        # We have said that inputs are subset of outputs, but here we don't necessarily assume that.

        newargs = {**kwargs, **res}
        if self.__check_termination_condition(kwargs, res):
            updated_args = {**newargs, **self.__accumulated_artifacts}
            ret_val = self.forward_artifacts(
                updated_args, {}, self.actor.get_output_artifacts()
            )
        else:
            args_to_pass = self.forward_artifacts(
                newargs, {}, self.actor.get_input_artifacts()
            )
            ret_val = self.act(**args_to_pass)

        self.gen_xml_elem(
            self.forward_artifacts(newargs, {}, self.actor.get_output_artifacts())
        )
        return ret_val

    def __type_check(self, a):
        """
        In iterative composition the type check only checks if there is at least one
        artifact which can be fed back to the actor.
        Other artifacts retain their initial values.
        """
        if not (a.get_input_artifacts().items() & a.get_output_artifacts().items()):
            raise CoVeriLangException(
                "Component \n {} cannot be composed with itself for iteration. "
                "There is no artifact to feed back in iteration. "
                "Type check failed.".format(a)
            )

    def __check_termination_condition(self, kwargs, res):
        # TODO at the moment I am assuming termination condition to be a common variable in the input and output.
        # Lets see if it is enough or we need to make it more expressive.
        return kwargs[self.terminationCondition] == res[self.terminationCondition]

    def __do_step_iteration_post_process(self, it_input, it_output):
        self.__collect_xml_elems(it_input)

        # Collect artifacts
        for k in self.__artifacts_to_accumulate:
            if k in self.__accumulated_artifacts:
                self.__accumulated_artifacts[k] = self.__accumulated_artifacts[k].join(
                    it_output[k]
                )
            else:
                self.__accumulated_artifacts[k] = it_output[k]

    def __collect_xml_elems(self, inputs):
        if self._iteration_count == 0:
            self.__initial_inputs = inputs
        it_elem = ElementTree.Element(
            "Iteration", {"count": str(self._iteration_count)}
        )
        it_elem.append(self.actor.xml_elem)
        self.__child_xml_elems.append(it_elem)
        self._iteration_count += 1

    def gen_xml_elem(self, outputs):
        super().gen_xml_elem(self.__initial_inputs, outputs)

        for i in self.__child_xml_elems:
            self.xml_elem.append(i)

    def __str__(self):
        return (
            "\nREPEAT"
            + self.__get_actor_type_str__()
            + "\n"
            + indent(str(self.actor), "\t")
        )


class Parallel(CompositeActor):
    def __init__(self, a1: Actor, a2: Actor):
        # TODO uncomment the type check. Removed it only for testing
        self.first = a1
        self.second = a2
        # TODO make a copy
        self._input_artifacts = {**a1.get_input_artifacts(), **a2.get_input_artifacts()}
        self._output_artifacts = {
            **a1.get_output_artifacts(),
            **a2.get_output_artifacts(),
        }

    def _act(self, **kwargs):
        # TODO It might not work as the arguments passed are more than required
        # TODO project the arguments as required
        kwargs1 = self.forward_artifacts(kwargs, {}, self.first.get_input_artifacts())
        kwargs2 = self.forward_artifacts(kwargs, {}, self.second.get_input_artifacts())
        res1 = self.first.act(**kwargs1)
        res2 = self.second.act(**kwargs2)
        ret_val = {**res1, **res2}
        self.gen_xml_elem(kwargs, ret_val)
        return ret_val

    def __type_check(self, a1, a2):
        # It just checks if the inputs and outputs are disjoint.
        # TODO commenting out the input type check as it was causing the problem.
        # Conflict cases: simple validated verifier, and meta val

        # TODO I wonder why is the following working. Inputs should be allowed to be shared.
        if dict_clash(a1.get_input_artifacts(), a2.get_input_artifacts()):
            raise CoVeriLangException(
                "Components \n {} \n and \n {} cannot be composed in parallel."
                "Input names clash. There is an input with the same name and"
                "different artifact type. Type check failed.".format(a1, a2)
            )
        if not a1.get_output_artifacts().keys().isdisjoint(a2.get_output_artifacts()):
            raise CoVeriLangException(
                "Components \n{} \nand \n {} cannot be composed in parallel."
                "Output names should be disjoint. Type check failed.".format(a1, a2)
            )

    def __str__(self):
        return (
            "\nPARALLEL"
            + self.__get_actor_type_str__()
            + "\n"
            + indent(str(self.first), "\t")
            + "\n"
            + indent(str(self.second), "\t")
        )

    def gen_xml_elem(self, inputs, outputs):
        super().gen_xml_elem(inputs, outputs)
        self.xml_elem.append(self.first.xml_elem)
        self.xml_elem.append(self.second.xml_elem)



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

import shutil
import sys
import argparse
import coveriteam.util as util
from coveriteam.interpreter.python_code_generator import generate_python_code
import logging
import coveriteam
from pathlib import Path
from coveriteam.remote_client import exec_remotely


class CoVeriTeam:
    def start(self, argv):
        self.config = self.create_argument_parser().parse_args(argv)
        str_to_prepend = ""

        if self.config.debug:
            FORMAT = "%(asctime)-15s %(levelname)s %(message)s"
            logging.basicConfig(level=util.CVT_DEBUG_LEVEL, format=FORMAT)

        if self.config.debug_deep:
            FORMAT = "%(asctime)-15s %(levelname)s %(message)s"
            logging.basicConfig(level=logging.DEBUG, format=FORMAT)

        if self.config.cache_dir:
            util.set_cache_directories(Path(self.config.cache_dir).resolve())
        else:
            util.set_cache_directories()

        if self.config.no_cache_update:
            util.set_cache_update(False)
        else:
            util.set_cache_update(True)

        if self.config.clean:
            # Remove both the archive and unzip directory
            shutil.rmtree(util.INSTALL_DIR)
            shutil.rmtree(util.ARCHIVE_DOWNLOAD_PATH)
            shutil.rmtree(util.TOOL_INFO_DOWNLOAD_PATH)
            util.create_cache_directories()

        if self.config.testtool:
            from coveriteam.language.atomicactor import AtomicActor

            a = AtomicActor(self.config.input_file)
            a.print_version()
            return

        if self.config.gettool:
            from coveriteam.language.atomicactor import AtomicActor

            a = AtomicActor(self.config.input_file)
            return

        if self.config.inputs:
            d = {}
            for k, v in self.config.inputs:
                if k in d:
                    # It should be either a string or a list of strings.
                    d[k] = [d[k]] + [v] if isinstance(d[k], str) else d[k] + [v]
                else:
                    d[k] = v
            for k in d:
                str_to_prepend += k + " = " + repr(d[k]) + ";\n"

            str_to_prepend += "from coveriteam.language.actor import Actor\n"
            data_model = (
                f'"{self.config.data_model}"' if self.config.data_model else "None"
            )
            str_to_prepend += "Actor.data_model = " + data_model + "\n"

            trusted = (
                f'"{self.config.trust_tool_info}"'
                if self.config.trust_tool_info
                else "False"
            )
            str_to_prepend += "Actor.trust_tool_info = " + trusted + "\n"

        if self.config.input_file:
            if self.config.remote:
                exec_remotely(self.config)
                return
            """"CVL file is provided."""
            generated_code = str_to_prepend + generate_python_code(
                self.config.input_file
            )
            if self.config.generate_code:
                print(generated_code)
            else:
                exec(generated_code, globals())  # noqa S102

    def create_argument_parser(self):
        """
        Create a parser for the command-line options.
        @return: an argparse.ArgumentParser instance
        """
        parser = argparse.ArgumentParser(
            prog="coveriteam",
            fromfile_prefix_chars="@",
            description="""Execute a program written in CoVeriLang.
               Command-line parameters can additionally be read from a file if file name prefixed with '@' is given as argument.
               """,
        )
        parser.add_argument(
            "--version", action="version", version=f"{coveriteam.__version__}"
        )

        parser.add_argument(
            "--get-tool",
            dest="gettool",
            action="store_true",
            default=False,
            help="Download the tool archive and initialize an atomic actor for the given YML file configuration.",
        )

        parser.add_argument(
            "--tool-info",
            dest="testtool",
            action="store_true",
            default=False,
            help="Test the given YML file configuration by executing the tool for version.",
        )

        parser.add_argument(
            "input_file",
            metavar="INPUT_FILE",
            help="The program written in CoVeriLang or a YML configuration if testing a tool.",
        )
        parser.add_argument(
            "--input",
            action="append",
            type=lambda kv: kv.split("="),
            dest="inputs",
            help="Inputs to the CoVeriLang program provided in the form of key=val.",
        )

        parser.add_argument(
            "--gen-code",
            dest="generate_code",
            action="store_true",
            default=False,
            help="Flag to generate python code from the cvl file.",
        )

        parser.add_argument(
            "--clean",
            dest="clean",
            action="store_true",
            default=False,
            help="Clean the tmp directory, which contains the extracted archives of the atomic actors.",
        )

        parser.add_argument(
            "--debug",
            dest="debug",
            action="store_true",
            default=False,
            help="Set the logging to debug for CoVeriTeam. Only logs debug (and higher) messages from CoVeriTeam",
        )

        parser.add_argument(
            "--debug-deep",
            dest="debug_deep",
            action="store_true",
            default=False,
            help="Set the logging to debug globally. Also logs debug (and higher) messages from BenchExec.",
        )

        parser.add_argument(
            "--cache-dir", metavar="CACHE_DIR", help="Path to the cache."
        )

        parser.add_argument(
            "--remote",
            dest="remote",
            action="store_true",
            default=False,
            help="Execute CoVeriTeam remotely.",
        )

        parser.add_argument(
            "--verbose",
            dest="verbose",
            action="store_true",
            default=False,
            help="Flag to print verbose output. At the moment supported only for remote execution.",
        )

        parser.add_argument(
            "--data-model",
            metavar="DATA_MODEL",
            help="Data model of the program under analysis."
            "Please choose ILP32 for 32 bit architecture, and LP64 for 64 bit."
            "Default value is ILP32.",
            type=str,
            choices={"ILP32", "LP64"},
        )

        parser.add_argument(
            "--trust-tool-info",
            dest="trust_tool_info",
            action="store_true",
            default=False,
            help="Trust the tool info modules. This will load them without a container.",
        )

        parser.add_argument(
            "--no-cache-update",
            dest="no_cache_update",
            action="store_true",
            default=False,
            help="Do not update the cache. Only use tools available in the cache. If tools are not available, then CoVeriTeam simply fails.",
        )
        return parser


def main(argv=None):
    args = argv or sys.argv
    CoVeriTeam().start(args[1:])


if __name__ == "__main__":
    main()



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from coveriteam.language.actor import TestGenerator
from coveriteam.language.atomicactor import AtomicActor
from coveriteam.language.artifact import CProgram, TestSpecification, TestSuite
from coveriteam.language import CoVeriLangException


class ProgramTester(TestGenerator, AtomicActor):
    _input_artifacts = {"program": CProgram, "test_spec": TestSpecification}
    _output_artifacts = {"test_suite": TestSuite}
    _result_files_patterns = ["**"]

    # It is a deliberate decision to not have the init function. We do not want anyone to
    # create instances of this class.

    def _prepare_args(self, program, test_spec):
        return [program.path, test_spec.path]

    def _extract_result(self):
        # extract result
        # Yes, test suite is directory.
        metadata_xmls = list(self.log_dir().glob("**/metadata.xml"))

        if not metadata_xmls:
            raise CoVeriLangException("No test suite found!!")
        elif len(metadata_xmls) != 1:
            msg = (
                "Found more than one metadata.xml found. Following is the list: \n"
                + "\n".join(map(str, metadata_xmls))
            )
            raise CoVeriLangException(msg)

        return {"test_suite": TestSuite(metadata_xmls[0].parent)}



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from coveriteam.interpreter.file_collector import collect_files
from coveriteam.language.actorconfig import ActorDefinitionLoader
from functools import reduce
import os
import sys
import requests
from pathlib import Path
import zipfile
from xml.etree import ElementTree

CVT_REMOTE_URL = "https://coveriteam.sosy-lab.org/execute"
# CVT_REMOTE_URL = "http://127.0.0.1:5000/execute"  # noqa E800
load_config = ActorDefinitionLoader.load_config
collect_included_files = ActorDefinitionLoader.collect_included_files


def get_included_configs(fs):
    ActorDefinitionLoader.add_constructor("!include", ActorDefinitionLoader.include)
    included_files = []
    for f in fs:
        if Path(f).suffix == ".yml":
            included_files += collect_included_files(load_config(f))
    return included_files


def expand_file_paths(paths):
    expanded = []
    for p in paths:
        if os.path.isabs(p):
            sys.exit(
                "There is a file with absolute path. Please use only relative paths."
            )
        if os.path.isdir(p):
            expanded += [str(x) for x in Path(p).rglob("*") if x.is_file()]
        else:
            expanded += [p]
    return expanded


def exec_remotely(config):
    inputs = config.inputs
    cvt_file = config.input_file
    global VERBOSE
    VERBOSE = config.verbose
    # Collect files
    fs = collect_files(cvt_file)
    # TODO check if inputs are actually file paths
    input_paths = [v for _, v in inputs]
    files_needed = [cvt_file] + fs + input_paths
    files_needed = expand_file_paths(files_needed)
    if reduce(lambda x, y: x or y, map(os.path.isabs, files_needed)):
        sys.exit("There is a file with absolute path. Please use only relative paths.")

    # collect all the required yml files.
    files_needed += get_included_configs(files_needed)

    # Preparing inputs
    cp = os.path.commonpath(list(map(os.path.abspath, files_needed)))
    working_dir = os.path.relpath(os.getcwd(), cp)
    json = {
        "coveriteam_inputs": inputs,
        "cvt_program": cvt_file,
        "working_directory": working_dir,
        "filenames": files_needed,
    }
    if config.data_model:
        json["additional_parameters"] = {"data_model": config.data_model}
    res = call_service(json)

    # This is deliberate. 'if res' checks if 200 < status code < 400
    if res is not None:
        show_result(res)


def call_service(data):
    try:
        response = requests.post(CVT_REMOTE_URL, json=data)
        return response
    except requests.exceptions.ConnectionError:
        print("The service disconnected unexpectedly!")
    return None


def show_result(response):
    if response.status_code == 504:
        print(
            "Received response code 504. "
            "Possibly your request couldn't be timely scheduled on the cloud."
        )
        return
    if response.status_code in (400, 500):
        print(response)
        show_error(response)
        return
    output_dir = Path("cvt-output")
    if not output_dir.exists():
        output_dir.mkdir()

    archive_path = output_dir / "cvt_remote_output.zip"

    with archive_path.open("wb") as f:
        f.write(response.content)

    with zipfile.ZipFile(archive_path, "r") as zf:
        zf.extractall(output_dir)
        speculative_output_path = (
            zf.filelist[1].filename.split("/")[0] if len(zf.filelist) > 1 else None
        )

    with (output_dir / "LOG").open("r") as log:
        print(
            "-------------------------------------------------------------------------\n"
            "The following log was produced by the execution of the CoVeriTeam program"
            "on the verifier cloud:"
        )
        print(log.read())
        print(
            "-------------------------------------------------------------------------\n"
            "END OF THE LOG FROM REMOTE EXECUTION"
        )
    if speculative_output_path:
        print(
            "\n\n The output produced by the CoVeriTeam program during execution can be found in the directory: cvt-output/%s"
            % speculative_output_path,
        )

    if VERBOSE:
        if speculative_output_path:
            exec_dir = output_dir.resolve() / Path(speculative_output_path)
            print_output_logs(exec_dir)
        else:
            print("Cannot figure out the names of tool output files.")


def print_output_logs(exec_dir):
    try:
        exec_trace = exec_dir / "execution_trace.xml"
        tool_output_paths = []
        if exec_trace.is_file():
            xml_root = ElementTree.parse(str(exec_trace))  # noqa S314
            for tool_output in xml_root.iter(tag="tool_output"):
                tool_output_paths += [exec_dir / Path(tool_output.text)]
        else:
            print("Couldn't find execution trace. Ouput logs maybe out of order.")
            tool_output_paths = exec_dir.glob("**/output.txt")

        for tool_output_path in tool_output_paths:
            actor_name = tool_output_path.parent.parent.name
            print(
                "\n\n---------------------------------------------------------------\n"
            )
            print("Output produced by the actor: %s \n\n" % actor_name)
            with tool_output_path.open("r") as log:
                print(log.read())
    except Exception:
        print("Something went wrong while printing output logs of tools!")


def show_error(response):
    if response.json().get("message", None):
        print("Server returned an error with  message: %s" % response.json()["message"])
    else:
        print(
            "Unexpected response from the server. Please contact the development team."
        )



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

import filecmp
import shutil
import uuid
import coveriteam.util as util
from pathlib import Path
import os
from benchexec.result import get_result_classification

VERDICT_TRUE = "TRUE"
VERDICT_FALSE = "FALSE"
VERDICT_UNKNOWN = "UNKNOWN"
VERDICT_ERROR = "ERROR"
VERDICT_TIMEOUT = "TIMEOUT"


class Artifact:
    def __init__(self, path: str):
        if path:
            self.path = str(Path(path).resolve())
        else:
            self.path = ""

    def __str__(self):
        return os.path.relpath(self.path, os.getcwd()) if self.path else ""

    def __repr__(self):
        return "'%s'" % os.path.relpath(self.path, os.getcwd()) if self.path else ""


class Joinable:
    def join(self, artifact2):
        """
        Definition of this function is to be provided by the child class.
        """
        pass


class BehaviorDescription(Artifact):
    pass


class Justification(Artifact):
    pass


class Verdict(Artifact):
    def __init__(self, verdict: str):
        self.verdict = verdict
        self.path = ""

    def __str__(self):
        return self.verdict

    def __repr__(self):
        return "'%s'" % self.verdict

    def __eq__(self, other):
        # Compare only true or false. We can see later if we need to compare the actual string.
        return get_result_classification(self.verdict) == get_result_classification(
            other
        )


class Specification(Artifact):
    pass


class Program(BehaviorDescription):
    # TODO move it to artifact at some point
    def __init__(self, path):
        self.multipath = False
        if not path:
            self.path = ""
        elif not isinstance(path, str) and isinstance(path, list):
            self.path = [str(Path(p).resolve()) for p in path]
            self.multipath = True
        else:
            self.path = str(Path(path).resolve())

    def __str__(self):
        if self.multipath:
            return str(
                [os.path.relpath(p, os.getcwd()) if p else "" for p in self.path]
            )
        return super().__str__()

    def __repr__(self):
        if self.multipath:
            return str(
                [
                    "'%s'" % os.path.relpath(p, os.getcwd()) if p else ""
                    for p in self.path
                ]
            )
        return super().__repr__()


class CProgram(Program):
    pass


class JavaProgram(Program):
    pass


class BehaviorSpecification(Specification):
    pass


class TestSpecification(Specification):
    pass


class Witness(Justification):
    pass


class ReachabilityWitness(Witness):
    pass


class Condition(Justification):
    pass


class TestGoal(Condition, Joinable):
    def join(self, other):
        assert type(self) is type(
            other
        ), "Cannot join %r and %r. Types are not same." % (self, other)
        if not self.path:
            return TestGoal(other.path)
        elif other.path:
            # If files are same then no need to compare
            if filecmp.cmp(self.path, other.path, False):
                return self
            joined = Path(self.path).parent / str(uuid.uuid4())
            with joined.open("w") as f:
                with open(self.path) as f1:
                    f.write(f1.read())
                with open(other.path) as f1:
                    f.write(f1.read())
            return TestGoal(joined)
        # When other does not exist
        return self

    def __eq__(self, other):
        if isinstance(other, TestGoal):
            if self.path and other.path:
                return filecmp.cmp(self.path, other.path, False)
            elif not (self.path or other.path):
                return True
            else:
                return False
        else:
            return False


class TestSuite(Justification):
    def join(self, other):
        assert type(self) is type(
            other
        ), "Cannot join %r and %r. Types are not same." % (self, other)
        if not self.path:
            self.path = other.path
        elif other.path:
            shutil.copytree(other.path, str(Path(self.path) / str(uuid.uuid4())))
        return self


class AtomicActorDefinition(Artifact):
    def __init__(self, actordef):
        self.actordef = actordef
        # Assumption: There should be a file present with the same name in our actor configs.
        if util.ACTOR_CONFIG_PATH:
            actordef_path = Path(util.ACTOR_CONFIG_PATH) / actordef / ".yml"
        else:
            this_path = Path(__file__).resolve()
            actordef_path = (
                this_path.parent.parent.parent / "actors" / (actordef + ".yml")
            )

        if actordef_path.exists():
            self.path = str(actordef_path)
        else:
            raise Exception("Actor %r could not be found!!" % actordef)

    def __str__(self):
        return self.actordef



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

from coveriteam.language.actor import Actor
from coveriteam.language.artifact import (
    Artifact,
    Joinable,
    Specification,
    TestSpecification,
)
from coveriteam.util import filter_dict, rename_dict
from coveriteam.util import INPUT_FILE_DIR


class UtilActor(Actor):
    def get_actor_kind():
        return "UtilityActor"


class IdentityActor(UtilActor):
    def name(self):
        return "Identity"

    def __init__(self, a: Actor):
        self._input_artifacts = a.get_input_artifacts()
        self._output_artifacts = a.get_output_artifacts()
        # TODO as a projection actor the outputs should be a subset of the inputs

    def _act(self, **kwargs):
        return filter_dict(kwargs, self.get_output_artifacts())


class CopyActor(UtilActor):
    def name(self):
        return "Identity"

    # TODO Should have an option for renaming
    def __init__(self, names):
        if type(names) is list:
            # The case where only a list of artifact to copy is provided.
            self._renaming_map = {}
            self._input_artifacts = {a: Artifact for a in names}
            self._output_artifacts = {a: Artifact for a in names}
        else:
            # The case where the artifacts should be copied and renamed.
            self._renaming_map = names
            self._input_artifacts = {a: Artifact for a in names.keys()}
            self._output_artifacts = {a: Artifact for a in names.values()}

    def _act(self, **kwargs):
        return rename_dict(kwargs, self._renaming_map)


# TODO delete by Jan 2021 if still commented.
"""
class SimpleExpressionActor(UtilActor):
    def name(self):
        return "SimpleExpressionActor"

    # TODO Should have an option for renaming
    def __init__(self, inputs, outputs, exp):
        self._input_artifacts = inputs
        self._output_artifacts = outputs
        self.exp = exp

    def _act(self, **kwargs):
        return eval(self.exp, globals(), {**kwargs})
"""


class Joiner(UtilActor):
    def name(self):
        return "Joiner"

    def __init__(self, t, ips, op):
        # TODO Ideally this type could be inferred in the composition
        assert issubclass(t, Joinable), "%r is not Joinable" % t
        self._input_artifacts = {ip: t for ip in ips}
        self._output_artifacts = {op: t}
        self.op = op

    def _act(self, **kwargs):
        ips = list(self._input_artifacts.keys())
        joined = kwargs[ips[0]].join(kwargs[ips[1]])
        return {self.op: joined}


class TestSpecToSpec(UtilActor):
    def name(self):
        return "TestSpecToSpec"

    def __init__(self):
        self._input_artifacts = {"test_spec": TestSpecification}
        self._output_artifacts = {"spec": Specification}

    def _act(self, **kwargs):
        # At the moment this works because we have only one test specification.
        # The day we add more this might stop working.
        property_path = INPUT_FILE_DIR + "specifications/unreach-call.prp"
        return {"spec": Specification(property_path)}


class SpecToTestSpec(UtilActor):
    def name(self):
        return "SpecToTestSpec"

    def __init__(self):
        self._input_artifacts = {"spec": Specification}
        self._output_artifacts = {"test_spec": TestSpecification}

    def _act(self, **kwargs):
        # At the moment this works because we have only one test specification.
        # The day we add more this might stop working.
        property_path = INPUT_FILE_DIR + "specifications/coverage-error-call.prp"
        return {"test_spec": TestSpecification(property_path)}



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import yaml
import benchexec
from pathlib import Path
from coveriteam.util import (
    is_url,
    download_if_needed,
    unzip,
    get_ARCHIVE_DOWNLOAD_PATH,
)
import coveriteam.util as util
import re
from coveriteam.language import CoVeriLangException
import os


class ActorDefinitionLoader(yaml.SafeLoader):
    def __init__(self, stream):
        self._root = Path(stream.name).parent
        super(ActorDefinitionLoader, self).__init__(stream)

    def include(self, node):
        filename = self._root / self.construct_scalar(node)
        with filename.open("r") as f:
            d = yaml.load(f, ActorDefinitionLoader)  # noqa S506
            return dict_merge(d, {"imported_file": filename})

    @staticmethod
    def load_config(path):
        with open(path, "r") as f:
            try:
                d = yaml.load(f, ActorDefinitionLoader)  # noqa S506
            except yaml.YAMLError as e:
                msg = "Actor config yaml file {} is invalid: {}".format(path, e)
                raise CoVeriLangException(msg, 203)

            return d

    @staticmethod
    def resolve_includes(d):
        # Check if "imports" exist
        imports = d.pop("imports", None)
        if not imports:
            return d

        if not isinstance(imports, list):
            imports = [imports]

        di = {}
        for i in imports:
            i.pop("imported_file", None)
            di = dict_merge(ActorDefinitionLoader.resolve_includes(i), di)
        return dict_merge(di, d)

    @staticmethod
    def collect_included_files(d):
        f = d.pop("imported_file", None)
        f = [] if f is None else [str(f)]
        # Check if "imports" exist
        imports = d.pop("imports", None)
        if not imports:
            return f

        if not isinstance(imports, list):
            imports = [imports]

        fs = []
        for i in imports:
            fs += ActorDefinitionLoader.collect_included_files(i)
        return fs + f


class ActorConfig:
    def __init__(self, path, version=None):

        ActorDefinitionLoader.add_constructor("!include", ActorDefinitionLoader.include)
        self.path = path
        self.get_actor_config()

        if self._config["format_version"] != "1.2":
            raise CoVeriLangException(
                f"Actor {path} version is not version 1.2 "
                + f"but version {self._config['format_version']}",
                200,
            )

        self.actor_name = self._config["actor_name"]
        self.reslim = self._config["resourcelimits"]

        if version is None:
            logging.warn(
                f"Warning: No version specified for actor {self.actor_name}. Taking the first version from the actor definition file.\n"
                + "You can specify the tool's version as a third parameter in the call to ActorFactory.create()."
            )

            if len(self._config["archives"]) <= 0:
                raise CoVeriLangException(
                    f"{self.actor_name} doesn't have any versions specified!"
                )
            self.version = str(self._config["archives"][0]["version"])
        else:
            self.version = str(version)

        tool_configs = [
            x for x in self._config["archives"] if x["version"] == self.version
        ]
        if len(tool_configs) > 1:
            raise CoVeriLangException(
                "There a multiple versions in the yaml file with the same name!", 2
            )

        tool_config = tool_configs[0]
        if tool_config is None:
            raise CoVeriLangException(
                f'{self.actor_name} doesn\'t recognize the requested version "{self.version}"!'
            )

        self.archive_location = tool_config["location"]

        if "options" in tool_config:
            self.options = tool_config["options"]
        else:
            logging.debug("No options override found, hence using global options.")
            self.options = self._config["options"]

        check_policy_compliance(self)
        self.archive_name = self.get_archive_name()
        # Keeping this path as str instead of Path because it is going to be used with string paths mostly.
        self.tool_dir = self.get_tool_installation_dir()
        if util.UPDATE_CACHE:
            self.__install_if_needed()
        self.__resolve_tool_info_module()

    def get_archive_name(self):
        archive_name = self.archive_location.rpartition("/")[2]
        if archive_name.endswith(".zip"):
            archive_name = archive_name.rpartition(".")[0]
        archive_name += "-" + self.version + ".zip"
        return archive_name

    def get_tool_installation_dir(self):
        # I think it is easier for debugging if the dir name starts with the archive name.
        # So, we take out the archive name and put it in the front instead of at the end.
        url_root = self.archive_location.rpartition("/")[0]
        if self.archive_name.endswith(".zip"):
            archive_name = self.archive_name.rpartition(".")[0]
        tool_dir_name = (
            archive_name + "-" + url_root.replace("://", "-").replace("/", "-")
        )
        return str(util.get_INSTALL_DIR() / tool_dir_name)

    def get_actor_config(self):
        self._config = ActorDefinitionLoader.load_config(self.path)
        self._config = ActorDefinitionLoader.resolve_includes(self._config)
        self.__check_actor_definition_integrity()
        self.__sanitize_yaml_dict()

    def __check_actor_definition_integrity(self):
        # check if the essential tags are present.
        # Essentiality of tags can be defined in a schema.
        essential_tags = [
            "toolinfo_module",
            "resourcelimits",
            "actor_name",
            "archives",
            "format_version",
        ]
        diff = essential_tags - self._config.keys()
        if diff:
            msg = (
                "The following tags are missing in the actor config YAML: "
                + self.path
                + "\n"
                + "\n".join(diff)
            )
            raise CoVeriLangException(msg, 200)

    def __sanitize_yaml_dict(self):
        # translate resource limits
        reslim = self._config.get("resourcelimits", None)
        if reslim:
            if reslim.get("memlimit"):
                reslim["memlimit"] = benchexec.util.parse_memory_value(
                    reslim.get("memlimit")
                )
            if reslim.get("timelimit"):
                reslim["timelimit"] = benchexec.util.parse_timespan_value(
                    reslim.get("timelimit")
                )
            self._config["resourcelimits"] = reslim

    def __install_if_needed(self):
        target_dir = Path(self.tool_dir)

        if not re.compile("^[A-Za-z0-9-._]+$").match(self.archive_name):
            raise CoVeriLangException(
                "Can't download tool because the version name contains invalid characters!",
                200,
            )

        archive_download_path = get_ARCHIVE_DOWNLOAD_PATH() / self.archive_name
        downloaded = download_if_needed(self.archive_location, archive_download_path)

        # Unzip only if downloaded or the tool directory does not exist.
        if downloaded or not target_dir.is_dir():
            print("Installing the actor: " + self.actor_name + "......")
            unzip(archive_download_path, target_dir)

    def __resolve_tool_info_module(self):
        """
        1. Check if it is a URL.
        2. If a URL then download it and save it to the TI cache.
        3. Infer the module name and return it.
        """
        # TODO the extraction of the tool info module name should be separated from downloading.
        ti = self._config["toolinfo_module"]
        if is_url(ti):
            filename = util.get_TOOL_INFO_DOWNLOAD_PATH() / ti.rpartition("/")[2]
            if util.UPDATE_CACHE:
                download_if_needed(ti, filename)
            ti = "." + filename.name

        if ti.endswith(".py"):
            ti = ti.rpartition(".")[0]

        self.tool_name = ti


def dict_merge(d1, d2):
    # Supposed to update but not overwrite. Instead update.
    for k in d2.keys():
        if k in d1.keys() and isinstance(d1[k], dict) and isinstance(d2[k], dict):
            if isinstance(d1[k], dict) and isinstance(d2[k], dict):
                d1[k] = dict_merge(d1[k], d2[k])
            elif isinstance(d1[k], dict) or isinstance(d2[k], dict):
                # TODO this could be and XOR
                # We raise an error when one of the values is a dict, but not the other.
                msg = "YAML file could not be parsed. Clash in the tag: %r" % k
                raise CoVeriLangException(msg, 201)
            else:
                d1[k] = d2[k]
        else:
            d1[k] = d2[k]

    return d1


def load_policy(policy_file):
    if not policy_file:
        if os.getenv("COVERITEAM_POLICY"):
            policy_file = os.getenv("COVERITEAM_POLICY")
        else:
            return {}

    with open(policy_file, "r") as f:
        try:
            return yaml.safe_load(f)
        except yaml.YAMLError as e:
            msg = "Failed to load policy from: {} Error is: {}".format(policy_file, e)
            raise CoVeriLangException(msg, 202)


def check_policy_compliance_allowed_locations(allowed_locations, archive_location):
    # Expression to match nothing.
    e = "a^"
    for loc in allowed_locations:
        e = "%s|(%s)" % (e, re.escape(loc))
    if not re.compile(e).match(archive_location):
        msg = "Not allowed to download from the url: %s" % archive_location
        raise CoVeriLangException(msg, 100)


def check_policy_compliance(ac, policy_file=None):
    policy = load_policy(policy_file)
    allowed_locations = policy.get("allowed_locations", None)
    if allowed_locations:
        check_policy_compliance_allowed_locations(
            allowed_locations, ac.archive_location
        )



#!/usr/bin/env python3

# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2020 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

import re
from setuptools import setup

with open("coveriteam/__init__.py") as inp:
    VERSION = re.search(r"^__version__\s*=\s*[\"\'](.*)[\"\']", inp.read(), re.M).group(
        1
    )

setup(
    name="CoVeriTeam",
    version=VERSION,
    author="Sudeep Kanav",
    description="CoVeriTeam: On-Demand Composition of Cooperative Verification Systems",
    url="https://gitlab.com/sosy-lab/software/coveriteam",
    packages=[
        "coveriteam",
        "coveriteam.interpreter",
        "coveriteam.language",
        "coveriteam.parser",
        "coveriteam.actors",
        "coveriteam.toolconfigs",
    ],
    package_data={"coveriteam": ["artifactlibrary/specifications/*.prp"]},
    install_requires=[
        "antlr4-python3-runtime>=4.8",
        "benchexec>=2.6",
        "requests>=2.24",
    ],
    setup_requires=[],
    tests_require=[],
    license="Apache-2.0",
    keywords="cooperative verification validation",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Testing",
    ],
    platforms=["Linux"],
    entry_points={"console_scripts": ["coveriteam = coveriteam.coveriteam:main"]},
)



# This file is part of CoVeriTeam, a tool for on-demand composition of cooperative verification systems:
# https://gitlab.com/sosy-lab/software/coveriteam
#
# SPDX-FileCopyrightText: 2021 Dirk Beyer <https://www.sosy-lab.org>
#
# SPDX-License-Identifier: Apache-2.0

# Generated from CoVeriLang.g4 by ANTLR 4.8
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .CoVeriLangParser import CoVeriLangParser
else:
    from CoVeriLangParser import CoVeriLangParser

# This class defines a complete generic visitor for a parse tree produced by CoVeriLangParser.


class CoVeriLangVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by CoVeriLangParser#program.
    def visitProgram(self, ctx: CoVeriLangParser.ProgramContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#fun_decl.
    def visitFun_decl(self, ctx: CoVeriLangParser.Fun_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#stmt_block.
    def visitStmt_block(self, ctx: CoVeriLangParser.Stmt_blockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#id_list.
    def visitId_list(self, ctx: CoVeriLangParser.Id_listContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#stmt.
    def visitStmt(self, ctx: CoVeriLangParser.StmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#spec_stmt.
    def visitSpec_stmt(self, ctx: CoVeriLangParser.Spec_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#PrintActor.
    def visitPrintActor(self, ctx: CoVeriLangParser.PrintActorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ExecuteActor.
    def visitExecuteActor(self, ctx: CoVeriLangParser.ExecuteActorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#return_stmt.
    def visitReturn_stmt(self, ctx: CoVeriLangParser.Return_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#SetActorName.
    def visitSetActorName(self, ctx: CoVeriLangParser.SetActorNameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#arg_map.
    def visitArg_map(self, ctx: CoVeriLangParser.Arg_mapContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#map_item_list.
    def visitMap_item_list(self, ctx: CoVeriLangParser.Map_item_listContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#map_item.
    def visitMap_item(self, ctx: CoVeriLangParser.Map_itemContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#assignable.
    def visitAssignable(self, ctx: CoVeriLangParser.AssignableContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Atomic.
    def visitAtomic(self, ctx: CoVeriLangParser.AtomicContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#FunCall.
    def visitFunCall(self, ctx: CoVeriLangParser.FunCallContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Utility.
    def visitUtility(self, ctx: CoVeriLangParser.UtilityContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Sequence.
    def visitSequence(self, ctx: CoVeriLangParser.SequenceContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ITE.
    def visitITE(self, ctx: CoVeriLangParser.ITEContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Iterative.
    def visitIterative(self, ctx: CoVeriLangParser.IterativeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Parallel.
    def visitParallel(self, ctx: CoVeriLangParser.ParallelContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ActorAlias.
    def visitActorAlias(self, ctx: CoVeriLangParser.ActorAliasContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Parenthesis.
    def visitParenthesis(self, ctx: CoVeriLangParser.ParenthesisContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Joiner.
    def visitJoiner(self, ctx: CoVeriLangParser.JoinerContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Copy.
    def visitCopy(self, ctx: CoVeriLangParser.CopyContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Rename.
    def visitRename(self, ctx: CoVeriLangParser.RenameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#TestSpecToSpec.
    def visitTestSpecToSpec(self, ctx: CoVeriLangParser.TestSpecToSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#SpecToTestSpec.
    def visitSpecToTestSpec(self, ctx: CoVeriLangParser.SpecToTestSpecContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Identity.
    def visitIdentity(self, ctx: CoVeriLangParser.IdentityContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#CreateArtifact.
    def visitCreateArtifact(self, ctx: CoVeriLangParser.CreateArtifactContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ArtifactAlias.
    def visitArtifactAlias(self, ctx: CoVeriLangParser.ArtifactAliasContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ArtifactFromMapItem.
    def visitArtifactFromMapItem(
        self, ctx: CoVeriLangParser.ArtifactFromMapItemContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#artifact_type.
    def visitArtifact_type(self, ctx: CoVeriLangParser.Artifact_typeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#actor_type.
    def visitActor_type(self, ctx: CoVeriLangParser.Actor_typeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ExpAlias.
    def visitExpAlias(self, ctx: CoVeriLangParser.ExpAliasContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#NotLogical.
    def visitNotLogical(self, ctx: CoVeriLangParser.NotLogicalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#InstanceOf.
    def visitInstanceOf(self, ctx: CoVeriLangParser.InstanceOfContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#ElementOf.
    def visitElementOf(self, ctx: CoVeriLangParser.ElementOfContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#BinaryLogical.
    def visitBinaryLogical(self, ctx: CoVeriLangParser.BinaryLogicalContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#Paren.
    def visitParen(self, ctx: CoVeriLangParser.ParenContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#verdict_list.
    def visitVerdict_list(self, ctx: CoVeriLangParser.Verdict_listContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#tc_exp.
    def visitTc_exp(self, ctx: CoVeriLangParser.Tc_expContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by CoVeriLangParser#quoted_ID.
    def visitQuoted_ID(self, ctx: CoVeriLangParser.Quoted_IDContext):
        return self.visitChildren(ctx)


del CoVeriLangParser
